import logging

logger = logging.getLogger(__name__)

import sys
import re
import os

from traceback import print_exc as xp
import pdb

from pip_stripper._baseutils import ppp, cpdb, rpdb
from pip_stripper.common import enforce_set_precedence


class ClassifierPip(object):
    """
    classifies pip packages names into buckets:
    - by hardcoded configuration in its entry in `pip-stripper.yaml:`, under `buckets`.
    - by bucket assignment per import directory in `ClassifierImport`
    
    """

    def __repr__(self):
        return self.__class__.__name__

    def __init__(self, mgr):
        try:
            self.mgr = mgr
            self.config = self.mgr.config.get(self.__class__.__name__)

            self.patre_splitline = re.compile(self.config["pattern_splitline"])

            self.di_bucket_association = self.config["buckets"]
            self.di_bucket = {}
            for k, v in self.di_bucket_association.items():
                self.di_bucket[k] = set()
                s_ = self.di_bucket_association[k] = set(v)
                try:
                    s_.remove("pass")
                except (KeyError,) as e:
                    pass

            self.di_freeze = {}

            self.bucket_precedence = self.config["bucket_precedence"]

            # remove entries in lower precedence positions that
            # exist higher up
            li = []
            for bucketname in self.bucket_precedence:
                li.append(self.di_bucket_association[bucketname])
            enforce_set_precedence(li)

            self.warnings = self.config["warnings"]

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def parse_requirement_line(self, line):
        try:
            packagename, version = self.patre_splitline.split(line)
            return packagename
        except ValueError:
            raise

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def run(self, packagetracker):

        try:

            di_packagename2pip = self.mgr.imp2pip
            all_imports = self.mgr.all_imports

            self.s_missing_imports = all_imports.copy()

            packagename = packagename_ = None
            pip2imp = self.mgr.pip2imp

            for pipname in self.mgr.all_pips:

                packagename = pip2imp.get(pipname, pipname)

                found = False
                for bucketname in self.bucket_precedence:
                    if pipname in self.di_bucket_association[bucketname]:
                        self.di_bucket[bucketname].add(pipname)
                        found = True
                        try:
                            self.s_missing_imports.remove(packagename)
                        except (KeyError,) as e:
                            pass
                        break

                if found:
                    continue

                bucketname = packagetracker.get_package(packagename)
                if bucketname:
                    self.di_bucket[bucketname].add(pipname)
                    try:
                        self.s_missing_imports.remove(packagename)
                    except (KeyError,) as e:
                        pass
                    continue

                self.di_bucket["unknown"].add(pipname)

            for import_ in self.s_missing_imports:
                self.warnings.append("missing import:%s" % (import_))

        except (Exception,) as e:
            if cpdb():
                ppp(dict(packagename=packagename, packagename_=packagename_))
                pdb.set_trace()
            raise



import logging

logger = logging.getLogger(__name__)

import sys
import os
from yaml import safe_load as yload

from traceback import print_exc as xp
import pdb

from pip_stripper._baseutils import ppp, cpdb, rpdb, sub_template
from pip_stripper.common import enforce_set_precedence


class Builder(object):
    def __init__(self, mgr):
        self.mgr = mgr
        self.fnp_input_classifier = self.mgr._get_fnp("scan")
        self.config = self.mgr.config.get(self.__class__.__name__)

    def process(self):
        try:

            with open(self.fnp_input_classifier) as fi:
                di_classifier = yload(fi)

            di_bucket = di_classifier["pips"]["buckets"]

            di_freeze = di_classifier["pips"]["freeze"]

            t_fn = self.config["t_filename_out"]
            for req, di in self.config["req_mapper"].items():
                fn_o = sub_template(t_fn, dict(req=req))
                fnp_o = os.path.join(self.mgr.workdir, fn_o)

                requirements = []

                for bucketname in di["buckets"]:
                    for pipname in di_bucket[bucketname]:
                        freeze_line = di_freeze[pipname]
                        requirements.append(freeze_line)

                requirements.sort()

                with open(fnp_o, "w") as fo:
                    for line in requirements:
                        fo.write("%s\n" % (line))

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise



import logging

logger = logging.getLogger(__name__)

import sys
import re
import os

from traceback import print_exc as xp
import pdb

from pip_stripper.yaml_commenter import Commenter

from pip_stripper._baseutils import (
    set_cpdb,
    set_rpdb,
    ppp,
    debugObject,
    cpdb,
    fill_template,
    rpdb,
    sub_template,
)

from yaml import safe_load as yload, dump


class ScanWriter(object):
    """
    imports:
      dev:
      prod:
      tests:

    pips:
      dev:
      prod:
      tests:
      workstation:
      unclassified:

    aliases:
    stdlib:

    """

    def __init__(self, mgr):
        self.mgr = mgr
        self.fnp_yaml = mgr._get_fnp("scan")

        self.fnp_yaml_comments = os.path.join(
            self.mgr._get_fnp("templatedir"), "yaml_comments.yaml"
        )

    def write(self):
        try:
            if rpdb():
                pdb.set_trace()

            pips_buckets = self.mgr.pip_classifier.di_bucket.copy()

            li_items = list(pips_buckets.items())

            for k, v in li_items:
                pips_buckets[k] = sorted(v)

                comment_key = k[:-1] + "_"
                comment_lookup = "comment_lookup_%s" % (k)

                pips_buckets[comment_key] = comment_lookup

            pips = dict(buckets=pips_buckets, freeze=self.mgr.all_freezes)

            warnings = self.mgr.pip_classifier.warnings

            di = self.di = dict(
                import_="comment_lookup_imports",
                imports=self.mgr.import_classifier.packagetracker.classify(),
                pips=pips,
                aliase_="comment_lookup_aliases",
                aliases=self.mgr.aliases,
                warnings=warnings,
            )

            if self.mgr.options.verbose:
                di_debug = dict(
                    stdlib=sorted(self.mgr.s_stdlib),
                    all_imports=sorted(self.mgr.all_imports),
                    all_pips=sorted(self.mgr.all_pips),
                )
                di["zzz_debug"] = di_debug

            fnp_tmp = self.mgr._get_fnp("tmp")
            with open(fnp_tmp, "w", encoding="utf-8") as fo:
                dump(self.di, fo, default_flow_style=False)

            commenter = Commenter(self.fnp_yaml_comments)

            commenter.comment(fnp_tmp, self.fnp_yaml)

            # raise NotImplementedError("write(%s)" % (locals()))
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise



import sys
import types
import logging

from string import Template

logger = logging.getLogger(__name__)


try:
    from cStringIO import StringIO  # 223ed
except (ImportError,) as e:
    from io import StringIO  # 223ed

try:
    basestring_ = basestring
except (NameError,) as e:
    basestring_ = str


def set_rpdb(rpdb, remove=False, recurse=True):
    if "--rpdb" in sys.argv:
        rpdb.enabled = True
        # howto- set cpdb on modules...
        if recurse:
            for module in sys.modules.values():
                try:
                    # this checks if it's likely to be cpdb, not something w same name
                    _ = module.rpdb.enabled
                    module.rpdb = rpdb
                except AttributeError:
                    pass

        if remove:
            sys.argv.remove("--rpdb")
        return rpdb.enabled
    return False


def set_cpdb(cpdb, remove=False, recurse=True, boolvalue=None):

    # !!!TODO!!!p4 - stop handling --pdb and consider it reserved

    single_use_flag = "--cpdb"

    flags = set(["--cpdbmany", "--cpdbonce", single_use_flag])
    args = set(sys.argv)

    # print("sys.argv#20", id(sys.argv), sys.argv, id(sys.argv))

    single_use_flag = "once" if single_use_flag in sys.argv else False

    found = single_use_flag or (args & flags)

    if boolvalue is None:
        boolvalue = getattr(cpdb, "enabled", False) or found

    if boolvalue:
        cpdb.enabled = boolvalue
        if recurse:
            for module in sys.modules.values():
                try:
                    # this checks if it's likely to be cpdb, not something w same name
                    _ = module.cpdb.enabled
                    module.cpdb = cpdb
                except AttributeError:
                    pass
    # import pdb
    # pdb.set_trace()
    if remove:
        # pdb.set_trace()
        for flag in flags:
            try:
                sys.argv.remove(flag)
            except (ValueError,) as e:
                pass
                # print("couldnt remove:%s:" % (flag) )
                # print("sys.argv#250", id(sys.argv), sys.argv)
    # print("sys.argv#29", id(sys.argv), sys.argv, id(sys.argv))


def ppp(o, header=None):
    if header is None:
        if isinstance(o, dict):
            header = "[%s %s...]" % (type(o), repr(o)[0:10])
        else:
            header = "[%s %s]" % (type(o), repr(o))

    # header = header or

    msg = debugObject(o, header)
    print(msg)


def debugObject(
    obj,
    header="",
    skip__=True,
    li_skipfield=[],
    indentlevel=0,
    sep="\n",
    li_skiptype=[types.MethodType],
    linefeed=1,
    truncate=60,
):

    if isinstance(obj, dict):
        di = obj
        return debugDict(
            di, header, li_skipfield, sep, indentlevel=indentlevel, linefeed=linefeed
        )
    elif isinstance(obj, list):
        return (
            header
            + "\n"
            + "\n".join(
                [
                    debugObject(
                        o, li_skipfield, sep, indentlevel=indentlevel, linefeed=linefeed
                    )
                    for o in obj[0:5]
                ]
            )
        )

    elif isinstance(obj, basestring_):
        return "\n%s:%s" % (header, obj)

    else:

        if not header and inspect.isclass(obj):
            header = "%s@%s" % (obj.__class__.__name__, obj.__module__)

        di = {}

        # properties can make a huge mess of things...
        li_skip_properties = []
        cls_ = getattr(obj, "__class__", None)
        if cls_:
            for k, v in vars(cls_).items():
                if isinstance(v, property):
                    li_skip_properties.append(k)

        for attrname in dir(obj):

            if attrname in li_skip_properties:
                continue

            if skip__ and attrname.startswith("__") and attrname.endswith("__"):
                continue
            try:
                attrval = di[attrname] = getattr(obj, attrname, "???")
            except Exception as e:
                attrval = "?"

            if type(attrval) in li_skiptype:
                li_skipfield.append(attrname)

        if not di:
            return "%s = %s" % (header, str(obj)[:500])

        return debugDict(
            di,
            header,
            li_skipfield,
            sep,
            indentlevel=indentlevel,
            linefeed=linefeed,
            truncate=truncate,
        )


# from repr import repr


def debugDict(
    dict,
    header="",
    li_skipfield=[],
    skip__=False,
    indentlevel=0,
    sep="\n",
    linefeed=1,
    truncate=60,
):

    buf = StringIO()

    try:
        li = list(dict.items())
    except AttributeError:
        return str(type(dict))

    try:
        li.sort()
    except (TypeError,) as e:
        logger.debug("debugDict:li.sort() error on %s" % (li))
        pass

    contents = " ={}" * (not li)  # huh?

    if truncate:
        msg = sep * linefeed + "%s%s %s\n" % (" " * indentlevel, header, repr(contents))
    else:
        msg = sep * linefeed + "%s%s %s\n" % (" " * indentlevel, header, str(contents))

    buf.write(msg)

    for k, v in li:
        sk = str(k)

        if skip__ and sk.startswith("__") and sk.endswith("__"):
            continue

        # make sure we dont leak secrets...
        if "secret" in sk.lower():
            continue

        if sk in li_skipfield:
            continue

        try:
            if truncate:
                buf.write("%s=%s%s" % (k, repr(v), sep))
            else:
                buf.write("%s=%s%s" % (k, str(v), sep))

        except UnicodeEncodeError:
            pass
        except AttributeError:
            buf.write("%s=%s%s" % (k, "?", sep))
        except Exception:
            pass

    return buf.getvalue()


def fill_template(tmpl, *args):  #!!!TODO!!! _baseutils
    try:
        return tmpl % Subber(*args)
    except Exception as e:
        logger.debug("tmpl:%s,args:%s" % (tmpl, args))
        raise


def sub_template(tmpl, *args):
    if isinstance(tmpl, basestring_):
        tmpl = Template(tmpl)

    try:
        return tmpl.substitute(Subber(*args))
    except Exception as e:
        logger.debug("tmpl:%s,args:%s" % (tmpl, args))
        raise


class Subber(object):
    def __init__(self, *args):
        self.li_arg = list(args)
        # raise NotImplementedError, "li_arg:%s" % (self.li_arg)
        self.di_res = {}

    @classmethod
    def factory(cls, obj, *args):

        li_sub = []

        for arg in args:
            if isinstance(arg, basestring_):
                li = arg.split(",")
                for arg2 in li:
                    arg2 = arg2.strip()
                    if arg2 == "self":
                        li_sub.append(obj)
                        continue
                    elif not "." in arg2:
                        sub = getattr(obj, arg2)
                        li_sub.append(sub)
                        continue
                    else:
                        raise NotImplementedError("!!!todo!!! rgetattr")
            else:
                li_sub.append(arg)
        return cls(*li_sub)

    def _get_from_args(self, key):
        """generic way to look for a key in the arg list"""

        for arg in self.li_arg:
            try:
                got = arg[key]
                return got
            except (KeyError):
                try:
                    # try getattr
                    got = getattr(arg, key)
                    return got
                except AttributeError:
                    continue

            except (AttributeError, TypeError):
                # no __getitem__, try getattr
                try:
                    got = getattr(arg, key)
                    return got
                except AttributeError:
                    continue

        try:
            raise KeyError(key)
        except Exception as e:
            raise

    def get(self, key, default=None):
        try:
            res = self._get_from_args(key)
            return res
        except KeyError:
            return default

    def __getitem__(self, keyname):

        """
        implement dictionary, attribute, func/method call lookup - a la Django Templates, but across
        a number of passed in arguments.

        if it's func/method call, then func(keyname, li_arg) is called (which includes the callable itself.

        """
        res = self._get_from_args(keyname)
        self.di_res[keyname] = res
        return res


class RescueDict(object):
    """fall through in case a fillTemplate does not find a key
       use sparingly as it covers up exceptions
    """

    def __init__(self, placeholder="?", template="%(key)s=%(placeholder)s"):
        self.placeholder = placeholder
        self.li_used = []
        self.template = template

    def __getitem__(self, key):
        self.li_used.append(key)
        return self.template % dict(key=key, placeholder=self.placeholder)


def cpdb(**kwds):
    if cpdb.enabled == "once":
        cpdb.enabled = False
        return True
    return cpdb.enabled


cpdb.enabled = False


def rpdb():
    return rpdb.enabled


rpdb.enabled = False



import logging

logger = logging.getLogger(__name__)

import os
import subprocess

from traceback import print_exc as xp
import pdb

from pip_stripper._baseutils import (
    ppp,
    debugObject,
    cpdb,
    fill_template,
    rpdb,
    sub_template,
)


def enforce_set_precedence(*sets):
    """
    given sets s1, s2, s3,  in that order
    s1 takes precedence over s2 and s3
    s2 takes precedence over s3

    -  any member already in s1 is removed from s2
    -  any member already in s1 or s2 is removed from s3
    """

    for ix, set in enumerate(sets):
        prunee = sets[ix]
        for prec_idx in range(0, ix):
            prunee -= sets[prec_idx]


class Command(object):
    def __init__(self, mgr, taskname, config):
        self.mgr = mgr
        self.taskname = taskname

        self.config = config
        self.append = self.config.get("append", False)
        self.stderr = ""

        self.mode = "a" if self.append else "w"

    def __repr__(self):
        return self.taskname

    def _subprocess(self, cmd, fnp_o, self_=None):
        try:

            fnp_stderr = self.mgr._get_fnp("log")
            with open(fnp_stderr, "a") as ferr:

                ferr.write("cmd: %s\nstderr begin:\n" % (cmd))

                with open(fnp_o, self.mode) as fo:
                    proc = subprocess.check_call(
                        cmd.split(), stdout=fo, stderr=ferr, cwd=self.mgr.workdir
                    )
                ferr.write("stderr end\n\n")

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    _subprocess_actual = _subprocess

    def run(self):
        try:
            t_cmd = self.config["cmdline"]  # .replace(r"\\","\\")
            t_fnp = os.path.join(self.mgr.workdir, self.config["filename"])

            fnp_log = "subprocess.log"

            cmd = sub_template(t_cmd, self, self.mgr.vars)

            fnp_o = sub_template(t_fnp, self, self.mgr.vars)
            self._subprocess(cmd=cmd, fnp_o=fnp_o, self_=self)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise



import logging

logger = logging.getLogger(__name__)


def cpdb(**kwds):
    cpdb.enabled = False


class Matcher(object):
    def __init__(self):

        self.pip = Side("pip", _variations_pip)
        self.imp = Side("imp", _variations_imp)

        self.di_pip_imp = dict()

    @classmethod
    def match_all(cls, mgr):
        try:
            inst = cls()
            for imp in mgr.all_imports:
                inst.imp.feed(imp)
            for pip in mgr.all_pips:
                inst.pip.feed(pip)

            inst.do_match()
            return inst.di_pip_imp

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def do_match(self):
        try:

            self.matches = self.pip.s_alias & self.imp.s_alias

            for match in self.matches:
                pip = self.pip.di_alias[match]
                imp = self.imp.di_alias[match]

                self.di_pip_imp[pip] = imp

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Side(object):
    def __repr__(self):
        return self.subject

    def __init__(self, subject, li_varfunc):
        self.subject = subject
        self.di_name = {}
        self.li_varfunc = li_varfunc
        self.s_alias = set()

        self.di_alias = {}

        self.di_duplicates = {}

    def get_aliases(self, name):
        try:

            def alias(name, func):
                try:
                    return func(name)
                except (Exception,) as e:
                    return None

            name = name.lower()
            res = set([i for i in [alias(name, f) for f in self.li_varfunc] if i])
            return res

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def feed(self, name):
        try:
            entry = self.di_name[name] = name

            aliases = self.get_aliases(name)

            # check for duplicates:
            dup = self.s_alias & aliases
            if dup:
                msg = "%s. alias duplicates:%s " % (self, dup)
                # raise NotImplementedError()
                logger.error(msg)

            self.s_alias |= aliases

            self.di_alias.update(**{k: entry for k in aliases})

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


_variations = [str]

_variations_pip = _variations + [lambda x: x.split("-")[1]]

_variations_imp = _variations + [lambda x: x.split("_")[1]]



import logging

logger = logging.getLogger(__name__)

import sys
import re
import os
from collections import defaultdict

from traceback import print_exc as xp
import pdb

from pip_stripper._baseutils import ppp, cpdb, rpdb
from pip_stripper.common import enforce_set_precedence


class ClassifierImport(object):
    def __repr__(self):
        return self.__class__.__name__

    def __init__(self, mgr):
        try:
            self.mgr = mgr

            self.config = self.mgr.config.get(self.__class__.__name__)

            self.buckets = []
            self.di_bucket = {}

            self.s_untracked = set(self.mgr.config.get("untracked", []))

            bucketnames_tracker = self.config["buckets"]["precedence"]

            self.packagetracker = PackageBucketTracker(self, bucketnames_tracker)

            self.workdir = self.mgr.workdir

            self.patre_splitline = re.compile(self.config["pattern_splitline"])

            for key, li_pattern in self.config["regex_dirs"].items():
                bucket = DirectoryPartitionerBucket(key)
                for pattern in li_pattern:
                    bucket.add(pattern)
                self.di_bucket[key] = bucket
                self.buckets.append(bucket)

            default_bucket_name = self.config["default_bucket"]
            self.default_bucket = self.di_bucket[default_bucket_name]

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def classify_filename(self, filename_import):
        try:
            for bucket in self.buckets:
                found = bucket.is_match(filename_import)
                if found:
                    return bucket

            return self.default_bucket

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def run(self):
        try:
            # raise NotImplementedError("%s.run(%s)" % (self, locals()))

            self.s_stdlib = self.mgr.s_stdlib

            for line in self.mgr.raw_imports:
                filebucket, packagename = self.parse(line)

                if packagename in self.s_stdlib:
                    if self.mgr.options.verbose:
                        logger.info("%s in std lib" % (packagename))
                    continue

                if packagename in self.s_untracked:
                    if self.mgr.options.verbose:
                        logger.info("%s in std lib" % (packagename))
                    continue

                self.packagetracker.add_import(packagename, filebucket.bucket.name)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def parse(self, line_in):

        """
        
        """
        # /Users/jluc/kds2/py2/bemyerp/lib/_baseutils.py:6:from functools import partial, wraps
        try:
            line = line_in.replace(self.workdir, "").rstrip()
            filename, import_ = self.patre_splitline.split(line)
            packagename = self.parse_import(import_)
            pythonfile = PythonFile.get(self, self, filename)

            return pythonfile, packagename

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def parse_import(self, import_: str) -> str:
        try:
            import_ = import_.strip()

            if import_.startswith("from ") or import_.startswith("import "):
                _packagename = import_.split()[1]
                packagename = _packagename.split(".")[0]
                return packagename

            raise NotImplementedError("parse_import(%s)" % (locals()))

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class PackageBucketTracker:
    def __init__(self, mgr, bucketnames):
        """
            this is a precedence mechanism
            if bucketnames : prod, dev
            then a packagename gets move from dev to prod
            if it was in dev but then gets used in prod

        """
        self.mgr = mgr
        self.bucketnames = bucketnames.copy()
        self.di_packagename = {}

        self.di_bucketindex = {}
        for ix, bucketname in enumerate(self.bucketnames):
            self.di_bucketindex[bucketname] = ix

    def get_package(self, packagename):
        return self.di_packagename.get(packagename)

    def add_import(self, packagename, bucketname):
        try:
            bucketname_prev = self.di_packagename.get(packagename)
            if not bucketname_prev:
                # first time, just put it in
                self.di_packagename[packagename] = bucketname
                return

            if bucketname == bucketname_prev:
                # same as before, nothing to do
                return

            index_new = self.di_bucketindex[bucketname]
            index_old = self.di_bucketindex[bucketname_prev]

            if index_new < index_old:
                # higher precendence for new (ex:  prod beats dev)
                self.di_packagename[packagename] = bucketname

            if rpdb():
                pdb.set_trace()
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def classify(self):
        di_result = defaultdict(list)

        for packagename, bucketname in self.di_packagename.items():
            di_result[bucketname].append(packagename)

        return dict(**di_result)

    def report(self):
        classification = self.classify()
        lines = []

        di_buckets = {}
        di = dict(package_imports=di_buckets)

        for bucketname in self.bucketnames:
            lines.append("%s:" % (bucketname))
            li = classification[bucketname]
            li.sort()
            # for import_ in li:
            #     lines.append("  %s" % (import_))
            di_buckets[bucketname] = li

            lines_extend(self.mgr, lines, bucketname, li)

        if self.mgr.fo:
            dump(di, self.mgr.fo, default_flow_style=False)
            self.mgr.fo.write("\n")

        print("\n".join(lines))


class DirectoryPartitionerBucket:
    def __init__(self, name):
        self.name = name
        self.li_pattern = []

    def __repr__(self):
        return "Bucket(%s)" % (self.name)

    def add(self, pattern):
        self.li_pattern.append(re.compile(pattern))

    def is_match(self, filename):
        try:
            for patre in self.li_pattern:
                hit = patre.search(filename)
                if hit:
                    return True
            return False
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class PythonFile:

    di_filename = {}

    def __repr__(self):
        return "%s in %s" % (self.filename, self.bucket)

    def __init__(self, filename, bucket):
        self.filename = filename
        self.bucket = bucket
        self.imports = []

    def add(self, import_):
        try:
            raise NotImplementedError()
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @classmethod
    def get(cls, mgr, directorypartitioner, filename):
        try:

            res = cls.di_filename.get(filename)
            if res:
                return res

            bucket = directorypartitioner.classify_filename(filename)
            res = cls.di_filename[filename] = cls(filename, bucket)
            return res

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise



import re
import pdb

from yaml import safe_load as yload, dump


from pip_stripper._baseutils import ppp, cpdb


class Commenter(object):
    """
    inject 1-line comments at precise locations in existing yaml files
    """

    @classmethod
    def markup(cls, key):
        return "comment_lookup_%s" % (key)

    patre = re.compile("""\s*[a-z_A-Z0-9]+\:\scomment_lookup_(?P<lookup>.+)$""")

    def __init__(self, fnp_lookup):
        self.fnp_lookup = fnp_lookup
        with open(fnp_lookup) as fi:
            self.comments = yload(fi)

    def lookup(self, key):
        return self.comments.get(key, "")

    def comment(self, fnp_in, fnp_out, mode="w"):
        try:
            with open(fnp_in) as fi, open(fnp_out, mode) as fo:
                for line in fi.readlines():
                    # pdb.set_trace()

                    hit = self.patre.search(line)
                    if hit:

                        indent = " " * (len(line) - len(line.lstrip()))
                        key = hit.groups()[0]
                        comment = self.lookup(key)
                        fo.write("\n%s#%s\n" % (indent, comment))
                    else:
                        fo.write(line)
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    __call_ = comment



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `pip_stripper` package."""

import os
import unittest
import pdb
import sys
import shutil
from glob import glob
from yaml import safe_load as yload, dump

import tempfile

from unittest.mock import patch

from traceback import print_exc as xp


from pip_stripper._pip_stripper import Main, __file__ as _mainfile, Command
from pip_stripper.matching import Matcher
from pip_stripper.common import enforce_set_precedence, Command
from pip_stripper.yaml_commenter import Commenter


undefined = object()


def build_args(**kwargs):
    pass


import pdb


def cpdb(**kwds):
    if cpdb.enabled == "once":
        cpdb.enabled = False
        return True
    return cpdb.enabled


cpdb.enabled = False


def rpdb(**kwds):
    return rpdb.enabled


rpdb.enabled = False

dn_cwd_start = os.getcwd()


# howto- figuring out the test script's
dn_testscripts = os.path.dirname(__file__)
if not dn_testscripts:
    fnp_script = os.path.join(dn_cwd_start, sys.argv[0])
    dn_testscripts = os.path.dirname(fnp_script)


from pip_stripper._baseutils import set_cpdb, ppp, debugObject, set_rpdb

FN_IMPORTS_GREP = "tmp.pip-stripper.imports.rpt"
FN_SCAN = "pip-stripper.scan.yaml"
BASEPREFIX = "tmp.pip_stripper"


DN_SAMPLE = os.path.join(dn_testscripts, "..", "sample")
DN_SEED01 = os.path.join(DN_SAMPLE, "tst.seedworkdir01/py")


class Base(unittest.TestCase):
    testdir = None

    dn_seed = DN_SEED01

    def write(self, msg):
        """use this to test stderr"""
        self.stderr = msg

    def set_testdirbase(self):
        try:
            Base.testdir_base = testdir_base = os.getenv("pip_stripper_testdir")
            if testdir_base:

                if not os.path.exists(testdir_base):
                    os.makedirs(testdir_base)
            else:
                Base.testdir_base = tempfile.mkdtemp(prefix="pip_stripper")
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def setUp(self):
        """Set up test fixtures, if any."""

        self.testdir_base = (
            getattr(self, "testdir_base", None) or self.set_testdirbase()
        )

        self.oldpwd = os.getcwd()

        if self.testdir:
            os.chdir(self.testdir)
        self.parser = Main.getOptParser()
        self.stderr = ""

    def tearDown(self):
        """Tear down test fixtures, if any."""

        oldpwd = getattr(self, "oldpwd", None)
        if oldpwd:
            os.chdir(oldpwd)

    def get_label(self):
        sself = "%s" % (self)
        method_, class_ = sself.split()
        class_ = class_.split(".")[1][:-1]

        return "%s.%s" % (class_, method_)


class Test_Bad_options(Base):
    """Tests for `pip_stripper` package."""

    testdir = dn_testscripts

    def test_000_badarg(self):

        try:
            with self.assertRaises(SystemExit):
                stderr = sys.stderr
                sys.stderr = self
                options = self.parser.parse_args(["-x"])
            sys.stderr = stderr
            self.assertTrue("unrecognized" in self.stderr)
        except (Exception, SystemExit) as e:
            if cpdb():
                pdb.set_trace()
            raise

    # @unittest.skipUnless(1 or False, "Need to fix")
    def test_001_noconfig(self):
        try:
            options = self.parser.parse_args([])
            mgr = Main(options)
            self.fail("should have complained about missing configuration")

        except (ValueError,) as e:
            self.assertTrue("configuration file" in str(e))
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_002_badconfig(self):
        try:
            options = self.parser.parse_args(["--config", "notexist.yaml"])
            mgr = Main(options)
            self.fail("expected IOError")
        except (IOError, FileNotFoundError) as e:
            pass
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class WriterMixin(object):

    baseprefix = BASEPREFIX
    prefix = ""
    _testdir = None

    @property
    def testdir(self):
        if self._testdir is None:
            # pdb.set_trace()
            testdir = os.path.join(Base.testdir_base, self.get_label())
            if not os.path.exists(testdir):
                os.makedirs(testdir)

            self.__class__._testdir = testdir

        return self._testdir

    def seed(self):
        try:

            dn_src = os.path.join(dn_testscripts, self.dn_seed)
            dn_tgt = os.path.join(self.testdir, "py")

            if not os.path.exists(dn_tgt):
                shutil.copytree(dn_src, dn_tgt)
            return dn_tgt

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def has_run(self, fn_marker=None):
        try:
            fn_marker = fn_marker or self.fn_marker
            fnp_marker = os.path.join(self.testdir, fn_marker)
            return os.path.exists(fnp_marker)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def get_file(self, fn, mode="r"):
        try:
            fnp = os.path.join(self.mgr.workdir, fn)
            return open(fnp, mode)
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class BaseCommand(WriterMixin, Base):
    def setUp(self):
        try:
            if self.__class__ in (BaseCommand,):
                return

            super(BaseCommand, self).setUp()
            options = self.parser.parse_args(["--init", "--workdir", self.testdir])

            self.mgr = Main(options)
            self.config = self.mgr.config["Command"]["tasks"][self.taskname]
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_run(self):
        if self.__class__ in (BaseCommand,):
            return

        try:
            command = Command(self.mgr, self.taskname, self.config)
            command.run()
            from time import sleep

            sleep(0.5)
            fnp = self.mgr._get_fnp(self.taskname)
            with open(fnp) as fi:
                data = fi.read()
                self.assertTrue(data, "%s was empty" % (fnp))

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class TestCommandFreeze(BaseCommand):
    """ this is mocked rather than hitting your real pip """

    taskname = "freeze"


class TestPip_Init(WriterMixin, Base):

    commandline = "--init"

    def get_options(self):
        return self.parser.parse_args(self.commandline.split())

    def setUp(self):
        super(TestPip_Init, self).setUp()
        self.options = self.get_options()
        self.mgr = Main(self.options)

    def test_001_init(self):
        try:
            print("self.testdir:%s" % (self.testdir))

            fnp = os.path.join(self.mgr.workdir, self.mgr.FN_CONFIG)
            with open(fnp) as fi:
                config = yload(fi)
            ppp(config)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class MockMgr(Main):
    pass


class TestParts(TestPip_Init):
    def setUp(self):
        super(TestParts, self).setUp()
        self.workdir = self.seed()
        self.mgr.__class__ = MockMgr

    def test_import_bucket(self):

        try:
            for exp, line in [
                ("tests", "py/tests/test_script01.py:4:import pyquery #dev-bound")
            ]:

                filebucket, packagename = self.mgr.import_classifier.parse(line)
                got = filebucket.bucket.name
                msg = "\nexp:%s<>%s:got for:\n  %s" % (exp, got, line)
                self.assertEqual(exp, got, msg)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class BasePip_Scan(WriterMixin, Base):
    pass

    fake_subprocess_payload = dict(
        freeze="""PyQt5==5.12
PyQt5-sip==4.19.14
pyquery==1.4.0
pytest==4.4.0
python-dateutil==2.8.0
python-editor==1.0.4
"""
    )

    # this is used to run scan only once
    fn_marker = FN_SCAN

    def setUp(self):
        super(BasePip_Scan, self).setUp()
        self.workdir = self.seed()

    def get_options(self):
        return self.parser.parse_args(["--init", "--workdir", self.workdir])

    def fake_subprocess(self, cmd, fnp_o, self_):
        try:
            if self_.taskname != "freeze":
                return self_._subprocess_actual(cmd, fnp_o, self_)

            with open(fnp_o, self_.mode) as fo:
                fo.write(self.fake_subprocess_payload["freeze"])

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_001_scan(self):
        try:

            with patch.object(
                Command, "_subprocess", side_effect=self.fake_subprocess
            ) as mock_method:

                options = self.get_options()
                self.mgr = Main(options)
                self.mgr.process()

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    exp_imports = []

    def test_002_greps(self):
        try:
            if not self.has_run():
                self.test_001_scan()

            with self.get_file(FN_IMPORTS_GREP) as fi:
                data = fi.read()

            for import_ in self.exp_imports:
                self.assertTrue(
                    import_ in data, "missing %s from %s" % (import_, FN_IMPORTS_GREP)
                )

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    exp_tests_requirements = []

    def test_0031_import_tracker(self):
        try:
            if not self.exp_tests_requirements:
                return

            if not self.has_run():
                self.test_001_scan()

            exp = "tests"

            for req in self.exp_tests_requirements:
                got = self.mgr.import_classifier.packagetracker.di_packagename[req]
                self.assertEqual(exp, got, "%s:exp:%s<>%s:got" % (req, exp, got))
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_0032_test_bucket(self):
        try:

            if not self.exp_tests_requirements:
                return

            if not self.has_run():
                self.test_001_scan()

            with self.get_file(FN_SCAN) as fi:
                data = yload(fi)

            got_all = data["pips"]["buckets"]["tests"]
            t_msg = "missing %s from %s.%s"

            for exp in self.exp_tests_requirements:

                found = self.mgr.import_classifier.packagetracker.di_packagename[exp]

                msg = t_msg % (exp, FN_SCAN, ":pips/tests/")
                self.assertTrue(exp in got_all, msg)
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_005_build(self):
        try:
            if not self.has_run():
                self.test_001_scan()

            options = self.parser.parse_args(["--build", "--workdir", self.workdir])
            self.build_mgr = Main(options)
            if rpdb():
                pdb.set_trace()
            self.build_mgr.process()

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class TestPip_Scan(BasePip_Scan):

    dn_seed = DN_SEED01

    # celery is a from-only, django is an import-only
    exp_imports = ["jinja2", "django", "celery"]
    exp_tests_requirements = ["pyquery"]


class TestPart(Base):
    pass


class TestMatchingBase(unittest.TestCase):

    li_pip = ["cx-Oracle", "requests", "python-dateutil", "xxx"]
    li_imp = ["dateutil", "cx_Oracle", "requests", "yyy"]

    exp = {
        "cx-Oracle": "cx_Oracle",
        "requests": "requests",
        "python-dateutil": "dateutil",
    }

    def __repr__(self):
        return self.__class__.__name__

    def test(self):
        try:
            self.matcher = Matcher()
            for name in self.li_pip:
                self.matcher.pip.feed(name)
            for name in self.li_imp:
                self.matcher.imp.feed(name)

            self.matcher.do_match()

            self.assertEqual(self.exp, self.matcher.di_pip_imp)

        except (Exception,) as e:
            ppp(self.matcher, "%s.matcher" % (self))
            ppp(self.matcher.pip, "pip")
            ppp(self.matcher.imp, "imp")
            if cpdb():
                pdb.set_trace()
            raise


class TestSetPrecedence(Base):
    def test_001_std(self):
        try:
            prod = set("prod")
            tests = set("tests")
            dev = set("dev")

            exp = dict(prod=set("prod"), tests=set("tes"), dev=set("v"))

            enforce_set_precedence(prod, tests, dev)

            got = dict(prod=prod, tests=tests, dev=dev)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_002_only_dev(self):
        try:
            prod = set("")
            tests = set("")
            dev = set("dev")

            exp = dict(prod=set(""), tests=set(""), dev=set("dev"))

            enforce_set_precedence(prod, tests, dev)

            got = dict(prod=prod, tests=tests, dev=dev)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class TestCommenter(WriterMixin, Base):

    fnp_lookup = os.path.join(DN_SAMPLE, "yaml_comments.yaml")

    def test_001_basic(self, *args, **kwargs):
        try:

            di = dict(
                taskname_="comment_lookup_scanner_tasknames",
                tasknames=["taskA", "taskB"],
                buckets=dict(
                    test0="comment_lookup_bucketnames", test1="test1", test2="test2"
                ),
            )
            fnp_tmp = os.path.join(self.testdir, "tmp.yaml")
            print("fnp_tmp:%s" % (fnp_tmp))
            with open(fnp_tmp, "w") as fo:
                dump(di, fo, default_flow_style=False)

            fnp_o = os.path.join(self.testdir, self.get_label())

            commenter = Commenter(self.fnp_lookup)

            commenter.comment(fnp_tmp, fnp_o)

            with open(fnp_o) as fi:
                print(fi.read())

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class TestMatchingJinja(TestMatchingBase):

    li_pip = ["Jinja2"]
    li_imp = ["jinja2"]

    exp = {"Jinja2": "jinja2"}


if __name__ == "__main__":
    # pdb.set_trace()
    set_cpdb(cpdb, remove=True)
    set_rpdb(rpdb, remove=True)

    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

# with open("README.md") as readme_file:
#     readme = readme_file.read()


from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["PyYAML"]

setup_requirements = []

test_requirements = []

setup(
    author="JL Peyret",
    author_email="jpeyret@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="strip out unnecessary pip packages from requirements",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    license="MIT license",
    package_data={
        # And include any *.msg files found in the 'hello' package, too:
        "templates": ["*"]
    },
    # long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="pip_stripper",
    name="pip_stripper",
    packages=find_packages(include=["pip_stripper"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/jpeyret/pip-stripper",
    version="0.1.7",
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""Main module."""

#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""doc for _main.py - """
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

import sys
import argparse
import re
import os
import subprocess
import distutils.sysconfig as sysconfig

from traceback import print_exc as xp
import pdb

from pip_stripper._baseutils import (
    set_cpdb,
    set_rpdb,
    ppp,
    debugObject,
    cpdb,
    fill_template,
    rpdb,
    sub_template,
)

from pip_stripper.writers import ScanWriter
from pip_stripper.matching import Matcher
from pip_stripper.pip import ClassifierPip
from pip_stripper.trackimports import ClassifierImport
from pip_stripper.common import Command
from pip_stripper.builder import Builder

from yaml import safe_load as yload, dump


if __name__ == "__main__":
    set_cpdb(cpdb, remove=True)


dn_script = os.path.dirname(__file__)
if not dn_script:
    fnp_script = os.path.join(dn_cwd_start, sys.argv[0])
    dn_script = os.path.dirname(fnp_script)


class Main(object):
    """ manages batch"""

    di_opt = {}

    def __repr__(self):
        return self.__class__.__name__

    def _get_fnp(self, subject):
        try:
            if subject == "templatedir":
                return os.path.join(dn_script, "templates")
            else:
                fn = self.config["vars"]["filenames"][subject]
            return os.path.join(self.workdir, fn)
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def __init__(self, options):

        try:
            self.options = options

            pwd = os.getcwd()
            self.workdir = self.options.workdir or pwd

            self.config = None

            fnp_config = self.options.config
            if self.options.init:
                fnp_config = self._initialize(fnp_config)

            if not fnp_config:
                for dn in [self.workdir, pwd]:
                    fnp_config = os.path.join(dn, self.FN_CONFIG)
                    try:
                        with open(fnp_config) as fi:
                            self.config = yload(fi)
                            break
                    except (IOError,) as e:
                        pass
                else:
                    raise ValueError("missing configuration file")

            else:
                with open(fnp_config) as fi:
                    self.config = yload(fi)

            #
            self.vars = dict()

            self.vars["scandir"] = self.workdir

            sectionname = "filenames"
            section = self.config["vars"][sectionname]

            for k, v in section.items():
                self.vars.update(**{"%s_%s" % (sectionname, k): v})

            self.import_classifier = ClassifierImport(self)

            self.scanwriter = ScanWriter(self)

            self.matcher = Matcher()

            self.builder = Builder(self)

        except (ValueError,) as e:
            raise
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def process(self):
        try:
            if self.options.scan:
                self.scanner = Scanner(self)
                self.scanner.run()
                self.import_classifier.run()

                for name in self.import_classifier.packagetracker.di_packagename:
                    self.matcher.imp.feed(name)

                pips = self.pip_classifier = ClassifierPip(self)

                for set_ in pips.di_bucket.values():
                    [self.matcher.pip.feed(name) for name in set_]

                pips.run(self.import_classifier.packagetracker)

                # for name in self.li_pip:

                self.scanwriter.write()

            if self.options.build:
                self.builder.process()

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    DN = os.path.dirname(__file__)
    FN_CONFIG = "pip-stripper.yaml"

    _s_stdlib = None

    @property
    def s_stdlib(self):

        if self._s_stdlib is None:

            self._s_stdlib = liststdlib()
            self._s_stdlib |= set(self.config.get("extra_stdlib", []))

        return self._s_stdlib

    _aliases = _imp2pip = None

    @property
    def imp2pip(self):
        if self._imp2pip is None:

            self._imp2pip = {v: k for k, v in self.aliases.items()}

        return self._imp2pip

    @property
    def aliases(self):
        if self._aliases is None:
            self._aliases = Matcher.match_all(self)

            # self._aliases = self.matcher.di_pip_imp.copy()
            self._aliases.update(**self.config.get("hardcoded_aliases", {}))

        return self._aliases

    pip2imp = aliases

    _raw_imports = None

    @property
    def raw_imports(self):
        if self._raw_imports is None:
            fnp = self._get_fnp("imports")
            with open(fnp) as fi:
                self._raw_imports = fi.readlines()
        return self._raw_imports

    _all_imports = None

    @property
    def all_imports(self):
        if self._all_imports is None:
            self._all_imports = set(
                self.import_classifier.packagetracker.di_packagename
            )

        return self._all_imports

    _all_freezes = None
    _all_pips = None

    @property
    def all_freezes(self):
        if self._all_freezes is None:
            # this triggers the pips which what populates
            # the freezes...
            self.all_pips
        return self._all_freezes

    @property
    def all_pips(self):
        if self._all_pips is None:
            self._all_pips = set()
            self._all_freezes = {}
            fnp = self._get_fnp("freeze")
            with open(fnp) as fi:
                for line in fi.readlines():
                    try:
                        packagename = self.pip_classifier.parse_requirement_line(line)
                    except (ValueError,) as e:
                        logger.warning("could not parse packagename on %s" % (line))
                        continue
                    self._all_pips.add(packagename)
                    self._all_freezes[packagename] = line.strip()

        return self._all_pips

    @classmethod
    def getOptParser(cls):

        parser = argparse.ArgumentParser()

        dest = "config"
        parser.add_argument(
            "--" + dest,
            action="store",
            help="%s - will look for %s in --workdir, current directory "
            % (dest, cls.FN_CONFIG),
        )

        dest = "scan"
        default = True
        parser.add_argument(
            "--" + dest,
            default=default,
            action="store_false",
            help="%s scan python files and classify packages [%s]" % (dest, default),
        )

        dest = "build"
        default = False
        parser.add_argument(
            "--" + dest,
            default=default,
            action="store_true",
            help="%s [%s]" % (dest, default),
        )

        dest = "init"
        parser.add_argument(
            "--" + dest,
            action="store_true",
            help="%s initialize the config file if it doesn't exist" % (dest),
        )

        dest = "workdir"
        parser.add_argument(
            "--" + dest,
            action="store",
            help="%s [defaults to config's value or current directory]" % (dest),
        )

        dest = "verbose"
        default = False
        parser.add_argument(
            "--" + dest,
            default=default,
            action="store_true",
            help="%s [%s]" % (dest, default),
        )

        # use snippet.optadd to expand choices.
        return parser

    def _initialize(self, fnp_config):

        try:
            fnp_config = os.path.join(self.workdir, self.FN_CONFIG)

            config = dict(coucou=1)

            # load the template file
            fnp_template = os.path.join(self.DN, "templates/pip-stripper.yaml")
            with open(fnp_template) as fi:
                tmpl = fi.read()

            seed = fill_template(tmpl, self)

            with open(fnp_config, "w") as fo:
                fo.write(seed)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


def liststdlib():
    """
    pretty grungy code, will need a rework
    """
    listed = set()

    std_lib = sysconfig.get_python_lib(standard_lib=True)
    for top, dirs, files in os.walk(std_lib):
        for nm in files:
            if nm != "__init__.py" and nm[-3:] == ".py":
                found = os.path.join(top, nm)[len(std_lib) + 1 : -3].replace("\\", ".")

                found = found.split("/")[0]
                listed.add(found)

    return listed


class Scanner(object):
    def __init__(self, mgr):
        self.mgr = mgr
        self.config = self.mgr.config.get(self.__class__.__name__)
        self.tasknames = self.config["tasknames"]

    def run(self):
        try:
            for taskname in self.tasknames:
                config = self.mgr.config.get("Command")["tasks"][taskname]

                command = Command(self.mgr, taskname, config)
                command.run()

            fnp_out = os.path.join(
                self.mgr.workdir, self.mgr.config["vars"]["filenames"]["liststdlib"]
            )
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


if __name__ == "__main__":

    set_cpdb(cpdb, remove=True)

    parser = Main.getOptParser()
    options = parser.parse_args()
    mgr = Main(options)
    mgr.process()
